'''
Computes 3S predictions (10% schedule + 90% k-NN).
'''
import operator
import sys
import os
import re
import csv
import json
import copy
PORT_FILE = '/public/xcsp_instances/new/scripts/solvers/cop/portfolios_cop'
INFOS = '/public/xcsp_instances/new/scripts/solvers/cop/infos'
BACKUP_SOLVER = 'g12cpx'
FEAT_SOLVER = 'gecode'
TIMEOUT = 900.0
DATASET = set(open('/public/xcsp_instances/new/dataset/cop/cop_final_minizinc'))
TEST_PATH = '/public/xcsp_instances/new/dataset/cop/train_test/test_'
FEAT_VECTOR = '/public/xcsp_instances/new/features/cop/feat_vector_rn'
SCHED_FILE = '/public/xcsp_instances/new/predictions/cop/3S/schedule'
with open(FEAT_VECTOR, 'r') as infile:
  feat_vector = json.load(infile)
with open(INFOS, 'r') as infile:
  infos = json.load(infile)
portfolios = csv.reader(open(PORT_FILE), delimiter='|', skipinitialspace = True)

def get_schedule(num_solvers, rep, fold):
  """
  read the file containing all the schedules and pick the correct one 
  according to the tested fold    
  """
  schedule = {}
  sched_reader = csv.reader(open(SCHED_FILE, 'r'), delimiter='|', skipinitialspace = True)
  for row in sched_reader:
    if int(row[0]) > int(num_solvers):
      break
    if int(row[0]) == int(num_solvers) and row[1] == rep and row[2] == fold: 
      schedule[row[3]] = float(row[4])
  return schedule

def compare_instances(instance1, instance2):
  """
  compute the distance between two instances
  """
  distance = 0.0
  for i in range(0, len(feat_vector[instance1])):
    d = float(feat_vector[instance1][i]) - float(feat_vector[instance2][i])
    distance += d * d
  return distance
        
def get_similar_instances(instance, training_instances, k = 10):
  """
  return the most k similar instances in the training set
  """
  distances = []
  for i in training_instances:
    distance = compare_instances(instance, i)
    distances.append((distance,i))
  distances.sort(key = lambda x : x[0])
  return distances[:k]

def get_best_solver(cluster, solvers):
  """
  Given a cluster of instances and a set of solvers, returns the best solver 
  (in solvers) according to the score (within the cluster), potentially using 
  resolution time for tie-breaking.
  """
  asp = {}
  ast = {}
  for solver in solvers:
    asp[solver] = 0.0
    ast[solver] = 0.0
  for (_,inst) in cluster:
    for solver in solvers:
      asp[solver] += infos[inst][solver]['score']
      ast[solver] += infos[inst][solver]['otime']
  maximum = max(asp.values())
  best_solvers = filter(lambda x : asp[x] == maximum, asp.keys())
  sorted_solvers = sorted(best_solvers, key = lambda x : ast[x])
  return sorted_solvers[0]

def compute_solving_times(schedule, solvers, instance, training_set):
  """
  Given an instance and the training set, compute the corresponding schedule, 
  simulate it and print the overcome on std output.
  """
  time = infos[instance][FEAT_SOLVER]['feat_time']
  similar_instances = get_similar_instances(instance, training_set)
  best_solver = get_best_solver(similar_instances, solvers)
  # Simulate portfolio execution and print the results.
  solvers_map = schedule
  if best_solver in solvers_map:
    del solvers_map[best_solver]
  schedule = solvers_map.items()
  schedule.append((best_solver, TIMEOUT))
  schedule.append((BACKUP_SOLVER, TIMEOUT))
  max_score = 0
  for (solver, timeout) in schedule:
    perf_time = infos[instance][solver]['otime']
    if solver == FEAT_SOLVER:
      perf_time -= infos[instance][FEAT_SOLVER]['fzn_time']
    if timeout >= perf_time:
      time += perf_time      
      if time <= TIMEOUT:
	score = infos[instance][solver]['score']
	if score == 1:
          print instance + "|1|" + str(time) + "|" + solver + "|" + str(schedule) + "|"
          return
        if score > max_score:
          max_score = score
      else:
        score = get_partial_score(instance, solver, TIMEOUT - (time - perf_time))
        if score > max_score:
          max_score = score
        print instance + "|" + str(max_score) + "|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"
        return
    else:
      time += timeout
      if time <= TIMEOUT:
	score = get_partial_score(instance, solver, timeout)
	if score == 1:
          print instance + "|1|" + str(time) + "|" + solver + "|" + str(schedule) + "|"
          return
        if score > max_score:
          max_score = score
      else:
        score = get_partial_score(instance, solver, TIMEOUT - (time - timeout))
        if score > max_score:
          max_score = score
        print instance + "|" + str(max_score) + "|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"
        return
  if time > TIMEOUT:
    time = TIMEOUT
  print instance + "|" + str(max_score) + "|" + str(time) + "|" + solver + "|" + str(schedule) + "|"

def get_partial_score(instance, solver, time, lb = 0.25, ub = 0.75):
  item = infos[instance][solver]
  vals = item['vals']
  values = [float(v) for (v, t) in vals.items() if float(t) <= time]
  if not values:
    if item['info'] == 'unsat' and item['otime'] <= time:
      return 1
    else:
      return 0
  if item['obj'] == 'min':
    val = min(values)
    if val < item['min_val']:
      return 0
  else:
    val = max(values)
    if val > item['max_val']:
      return 0
  score = compute_score(
    item['info'], 
    item['min_val'], 
    item['max_val'], 
    item['obj'], 
    val,
    time,
    item['otime'],
    lb, 
    ub
  )
  if score < 0:
    return 0
  return score  
  
def compute_score(info, min_val, max_val, obj, val, time, otime, lb = 0.25, ub = 0.75):
  if info == "unknown":
    return 0
  if info != "sat" and otime <= time:
    return 1
  if min_val == max_val:
    return ub
  s = ((val - min_val) * (ub - lb)) / (max_val - min_val)
  assert ub - s <= ub or s + lb <= ubs
  if obj == "min":
    return ub - s
  if obj == "max":
    return s + lb
  print "Error!"
  system.exit(1)

def compute_predictions(solv, rep, fold, portfolio):
  """
  Predict a schedule of solvers for each instance of the test set.
  """
  # compute the instances of the train/test set.
  test_set = set(open(TEST_PATH + 'rep' + rep + '_fold' + fold))
  train_set = list(DATASET - test_set)
  train_set = [inst[0 : len(inst) - 1] for inst in train_set]
  #print len(test_set), len(train_set)
  #get the static schedule
  schedule = get_schedule(solv, rep, fold)
  for i in test_set:
    compute_solving_times(copy.copy(schedule), portfolio, i[0 : len(i) - 1], train_set)
        
def main(args):
  n_solver = args[0]
  rep      = args[1]
  fold     = args[2]
  for row in portfolios:
    if row[0] == n_solver:
      portfolio = eval(row[1])
      break
  compute_predictions(n_solver, rep, fold, portfolio)

if __name__ == '__main__':
    main(sys.argv[1:])
