'''
Computes the 3S fixed schedule of solvers within a time-out of 
C seconds, using precomputed coefficient matrices and headers.
'''

import sys
import os
import csv
import json
from gurobipy import *
from time import clock

with open('infos', 'r') as infile:
  infos = json.load(infile)
in_path = '/home/roberto/3S/'
C = 90.0 #180.0

def get_partial_score(instance, solver, time, lb = 0.25, ub = 0.75):
  item = infos[instance][solver]
  vals = item['vals']
  values = [float(v) for (v, t) in vals.items() if float(t) <= time]
  if not values:
    if item['info'] == 'unsat' and item['otime'] <= time:
      return 1
    else:
      return 0
  if item['obj'] == 'min':
    val = min(values)
    if val < item['min_val']:
      return 0
  else:
    val = max(values)
    if val > item['max_val']:
      return 0
  score = compute_score(
    item['info'], 
    item['min_val'], 
    item['max_val'], 
    item['obj'], 
    val,
    time,
    item['otime'],
    lb, 
    ub
  )
  if score < 0:
    return 0
  return score  
  
def compute_score(info, min_val, max_val, obj, val, time, otime, lb = 0.25, ub = 0.75):
  if info == "unknown":
    return 0
  if info != "sat" and otime <= time:
    return 1
  if min_val == max_val:
    return ub
  s = ((val - min_val) * (ub - lb)) / (max_val - min_val)
  assert ub - s <= ub or s + lb <= ubs
  if obj == "min":
    return ub - s
  if obj == "max":
    return s + lb
  print "Error!"
  system.exit(1) 
  
def main(args):
  for n_solver in range(10, 13): #2
    for rep in range(1, 6):
      for fold in range(1, 6):
        out_file = csv.writer(open(in_path + 'schedule', 'ab'), delimiter='|', skipinitialspace = True)
        #n_solver = args[0]
        #rep = args[1]
        #fold = args[2]
	start = clock()
	name = 'train_rep' + str(rep) + '_fold' + str(fold)
	print 'Building model ' + name + '...'
	model = Model(name)
	model.setParam('OutputFlag', False)
        path = in_path + str(n_solver) + '_solvers/'
	# coeff is a coefficient matrix of size (N + 1) * M.
	coeff  = csv.reader(open(path + 'coeff_'  + name, "rb"), delimiter=" ", skipinitialspace = True)
	# header is a row of length M which contains the names of the variables.
	header = csv.reader(open(path + 'header_' + name, "rb"), delimiter=" ", skipinitialspace = True)
	
	print 'Adding binary variables to the model...'
	coeff_list = list(coeff)
	N = len(coeff_list) - 1
	header_list = list(header)[0]
	# The last element of the header (and the last column of the coefficient matrix) is actually a blank space.
	M = len(header_list) - 1
	x = [Var] * M
	for i in range(0, N):
	  # header_list[i] (i.e. the name of the instance) is too long.
	  x[i] = model.addVar(vtype = GRB.BINARY, name = 'y_' + str(i))
	for j in range(N, M):
	  x[j] = model.addVar(vtype = GRB.BINARY, name = header_list[j])
	model.update()
	
	print 'Adding contraints to the model...'
	for i in range(0, N):
	  row_vars = []
	  for j in range(N, M):
	    if int(coeff_list[i][j]) == 1:
	      row_vars.append(x[j])
	  model.addConstr(x[i] + quicksum(row_vars) >= 1)
	
	v1 = x[N].varName
	k1 = v1.index('_')
	s1 = v1[0 : k1]
	solver_vars = [x[N]]
	for j in range(N + 1, M):
	  v2 = x[j].varName
	  k2 = v2.index('_')
	  s2 = v2[0 : k2]
	  if s1 == s2:
	    solver_vars.append(x[j])
	  else:
	    model.addConstr(quicksum(solver_vars) <= 1)
	    solver_vars = [x[j]]
	    s1 = s2
	model.addConstr(quicksum(solver_vars) <= 1)
	exprs  = []
	for k in range(N, M):
	  t = coeff_list[N][k]
	  exprs.append(LinExpr(float(t), x[k]))
	model.addConstr(quicksum(exprs) <= C)
	
	print 'Adding objective function to the model...'
	C1 = -(C * C)
	C2 = C
	C3 = -(1 / C)
	exprs = []
	for i in range(0, N):
	  inst = header_list[i]
	  for j in range(N, M):
	    solver_time = header_list[j]
	    k = solver_time.index('_')
	    solver = solver_time[0 : k]
	    time = float(solver_time[k + 1 : len(solver_time)])
	    score = get_partial_score(inst, solver, time)
	    exprs.append(LinExpr(C2 * score, x[j]))
	    exprs.append(LinExpr(C3 * time,  x[j]))
	for i in range(0, N):
	  exprs.append(LinExpr(C1, x[i]))
	model.setObjective(quicksum(exprs), GRB.MAXIMIZE)
	
	print 'Solving the model...'
	model.optimize()
	elapsed = (clock() - start)
	print 'Objective value:', model.objVal
	print 'Schedule results:'
	sol = model.getVars()
	print 'Solution length: ' + str(len(sol))
	k = 0
	for i in range(0, N):
	  if sol[i].x == 1:
	    k += 1
	    #print "Solver schedule can not solve instance: " + header_list[i]
	print "Total no. of instances that solver can't solve: " + str(k) + "/" + str(N)
	
	k = 0
	for j in range(N, M):
	  if round(sol[j].x) == 1:
	    solver_time = sol[j].varName
	    k = solver_time.index('_')
	    solver = solver_time[0 : k]
	    time = solver_time[k + 1 : len(solver_time)]
	    print "Solver schedule will run: " + solver_time
	    line = [n_solver, rep, fold, solver, time, elapsed]
	    out_file.writerow(line)
            #print n_solver + '|' + rep + '|' + fold + '|' + solver + '|' + str(time) + '|' + str(elapsed)
	  
if __name__ == '__main__':
    main(sys.argv[1:])
