"""
Compute SUNNY schedules.
"""
import sys
import csv
import json
from combinations import *
from math import sqrt

PORT_FILE = '/public/xcsp_instances/new/scripts/solvers/cop/portfolios_cop'
INFOS = '/public/xcsp_instances/new/scripts/solvers/cop/infos'
BACKUP_SOLVER = 'g12cpx'
FEAT_SOLVER = 'gecode'
TIMEOUT = 900.0
DATASET = set(open('/public/xcsp_instances/new/dataset/cop/cop_final_minizinc'))
TEST_PATH = '/public/xcsp_instances/new/dataset/cop/train_test/test_'
FEAT_VECTOR = '/public/xcsp_instances/new/features/cop/feat_vector_rn'
with open(FEAT_VECTOR, 'r') as infile:
  feat_vector = json.load(infile)
with open(INFOS, 'r') as infile:
  infos = json.load(infile)
portfolios = csv.reader(open(PORT_FILE), delimiter='|', skipinitialspace = True)
PORTFOLIO = {}
SOLV = 0

def compare_instances(instance1, instance2):
  """
  Compute the Euclidean distance between two instances.
  """
  distance = 0.0
  for i in range(0, len(feat_vector[instance1])):
    d = float(feat_vector[instance1][i]) - float(feat_vector[instance2][i])
    distance += d * d
  return sqrt(distance)
        
def get_similar_instances(instance, training_instances, k = 10):
  """
  Return the most k similar instances in the training set.
  """
  distances = []
  for i in training_instances:
    distance = compare_instances(instance, i)
    distances.append((distance,i))
  distances.sort(key = lambda x : x[0])
  return distances[:k]

def get_schedule(instance, neighborhood):
  """
  Given the neighborhood of a given instance, return the corresponding 
  SUNNY schedule.
  """
  scores = {}
  times  = {}
  for solver in PORTFOLIO:
    scores[solver] = []
    times[solver] = 0.0
  for (_,inst) in neighborhood:
    for solver in PORTFOLIO:
      scores[solver].append(infos[inst][solver]['score']) 
      times[solver] += infos[inst][solver]['otime']
  max_score = 0
  min_time = 1000000000
  best_portfolio = []
  # Select the best sub-portfolio.
  for i in range(1, SOLV + 1):
    for j in range(0, binom(SOLV, i)):
      score = 0
      solving_time = 0
      sub_portfolio = get_subset(j, i, PORTFOLIO)
      port_scores = dict([(s, l) for s, l in scores.items() if s in sub_portfolio])
      for k in range(0, K):
	score += max([inst[k] for inst in port_scores.values()])
      solving_time = sum([times[solver] for solver in sub_portfolio])
      if score >  max_score or \
        (score == max_score and solving_time < min_time):
          min_time  = solving_time
          max_score = score
          best_portfolio = sub_portfolio
  # n is the number of instances solved by each solver plus the instances 
  # that no solver can solver.
  n = sum([sum(scores[s]) for s in best_portfolio]) + (K - max_score)
  schedule = {}
  # compute the schedule and sort it by number of solved instances.
  for solver in best_portfolio:
    ns = sum(scores[solver])
    if ns == 0:
      continue
    schedule[solver] = TIMEOUT / n * ns
  tot_time = sum(schedule.values())
  if tot_time < TIMEOUT:
    if BACKUP_SOLVER in schedule.keys():
      schedule[BACKUP_SOLVER] += TIMEOUT - tot_time
    else:
      schedule[BACKUP_SOLVER]  = TIMEOUT - tot_time
  sorted_schedule = sorted(schedule.items(), key = lambda x: times[x[0]])
  #sorted_schedule = sorted(schedule.items(), key = lambda x: sum(scores[x[0]]), reverse = True)
  #print sorted_schedule
  if sum([t for (s,t) in sorted_schedule]) > TIMEOUT + 0.1:
    sys.stderr.write(instance + '|')
    sys.stderr.write(str(SOLV) + '|')
    sys.stderr.write(str(REP) + '|')
    sys.stderr.write(str(FOLD) + '|')
    sys.stderr.write(str(K) + '|')
    sys.stderr.write(str(schedule) + '|')
    sys.stderr.write(str(sum([t for (s,t) in sorted_schedule])) + '\n')
  return sorted_schedule

def compute_inst_schedule(instance, training_set):
  """
  Given an instance and the training set, compute the corresponding schedule, 
  simulate it and print the overcome on std output.
  """
  time = infos[instance][FEAT_SOLVER]['feat_time']
  similar_instances = get_similar_instances(instance, training_set, K)
  schedule = get_schedule(instance, similar_instances)
  # Simulate portfolio execution and print the results.
  max_score = 0
  for (solver, timeout) in schedule:
    perf_time = infos[instance][solver]['otime']
    if solver == FEAT_SOLVER:
      perf_time -= infos[instance][FEAT_SOLVER]['fzn_time']
    if timeout >= perf_time:
      time += perf_time      
      if time <= TIMEOUT:
	score = infos[instance][solver]['score']
	if score == 1:
          print instance + "|1|" + str(time) + "|" + solver + "|" + str(schedule) + "|"
          return
        if score > max_score:
          max_score = score
      else:
        score = get_partial_score(instance, solver, TIMEOUT - (time - perf_time))
        if score > max_score:
          max_score = score
        print instance + "|" + str(max_score) + "|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"
        return
    else:
      time += timeout
      if time <= TIMEOUT:
	score = get_partial_score(instance, solver, timeout)
	if score == 1:
          print instance + "|1|" + str(time) + "|" + solver + "|" + str(schedule) + "|"
          return
        if score > max_score:
          max_score = score
      else:
        score = get_partial_score(instance, solver, TIMEOUT - (time - timeout))
        if score > max_score:
          max_score = score
        print instance + "|" + str(max_score) + "|" + str(TIMEOUT) + "|" + solver + "|" + str(schedule) + "|"
        return
  if time > TIMEOUT:
    time = TIMEOUT
  print instance + "|" + str(max_score) + "|" + str(time) + "|" + solver + "|" + str(schedule) + "|"

def get_partial_score(instance, solver, time, lb = 0.25, ub = 0.75):
  item = infos[instance][solver]
  vals = item['vals']
  values = [float(v) for (v, t) in vals.items() if float(t) <= time]
  if not values:
    if item['info'] == 'unsat' and item['otime'] <= time:
      return 1
    else:
      return 0
  if item['obj'] == 'min':
    val = min(values)
    if val < item['min_val']:
      return 0
  else:
    val = max(values)
    if val > item['max_val']:
      return 0
  score = compute_score(
    item['info'], 
    item['min_val'], 
    item['max_val'], 
    item['obj'], 
    val,
    time,
    item['otime'],
    lb, 
    ub
  )
  if score < 0:
    return 0
  return score  
  
def compute_score(info, min_val, max_val, obj, val, time, otime, lb = 0.25, ub = 0.75):
  if info == "unknown":
    return 0
  if info != "sat" and otime <= time:
    return 1
  if min_val == max_val:
    return ub
  s = ((val - min_val) * (ub - lb)) / (max_val - min_val)
  assert ub - s <= ub or s + lb <= ubs
  if obj == "min":
    return ub - s
  if obj == "max":
    return s + lb
  print "Error!"
  system.exit(1)
  
def compute_predictions():
  """
  Predict a schedule of solvers for each instance of the test set.
  """
  # compute the instances of the train/test set.
  test_set = set(open(TEST_PATH + 'rep' + REP + '_fold' + FOLD))
  train_set = list(DATASET - test_set)
  train_set = [inst[0 : len(inst) - 1] for inst in train_set]
  #print len(test_set), len(train_set)
  for inst in test_set:
    compute_inst_schedule(inst[0 : -1], train_set)

def main(args):
  """
  Given the number of solvers, repetition, fold and the k parameter of k-NN 
  print on std output the overcome of SUNNY simulation for every instance of 
  the test set.
  """
  global SOLV, REP, FOLD, K, PORTFOLIO
  SOLV  = args[0]
  REP   = args[1]
  FOLD  = args[2]
  K     = int(args[3])
  # Look for the portfolio of size SOLV.
  for row in portfolios:
    if row[0] == SOLV:
      PORTFOLIO = eval(row[1])
      break
  SOLV = int(SOLV)
  compute_predictions()

if __name__ == '__main__':
    main(sys.argv[1:])
